@@ -32,15 +32,23 @@ import subprocess
 import tempfile
 import shutil
 import urllib.parse
+from typing import Set
 
 import adtlog
 import VirtSubproc
+from testdesc import Unsupported
 
 
 timeouts = {'short': 100, 'copy': 300, 'install': 3000, 'test': 10000,
             'build': 100000}
 
 
+# When running installed, this is /usr/share/autopkgtest.
+# When running uninstalled, this is the source tree.
+# Either way, it has a setup-commands subdirectory.
+PKGDATADIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
+
+
 class Testbed:
     def __init__(self, vserver_argv, output_dir, user,
                  setup_commands=[], setup_commands_boot=[], add_apt_pockets=[],
@@ -58,7 +66,6 @@ class Testbed:
         self.output_dir = output_dir
         self.shared_downtmp = None  # testbed's downtmp on the host, if supported
         self.vserver_argv = vserver_argv
-        self.install_tmp_env = []
         self.user = user
         self.setup_commands = setup_commands
         self.setup_commands_boot = setup_commands_boot
@@ -82,6 +89,10 @@ class Testbed:
         self.nproc = None
         self.cpu_model = None
         self.cpu_flags = None
+        self._created_user = False
+        self._tried_debug = False
+        # False if not tried, True if successful, error message if not
+        self._tried_provide_sudo = False
 
         try:
             self.devnull = subprocess.DEVNULL
@@ -96,16 +107,17 @@ class Testbed:
 
     def start(self):
         # log date at least once; to ease finding it
-        adtlog.info('starting date: %s' % time.strftime('%Y-%m-%d'))
+        adtlog.info('starting date and time: %s' % time.strftime('%Y-%m-%d %H:%M:%S%z'))
 
         # are we running from a checkout?
-        root_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
+        root_dir = PKGDATADIR
         if os.path.exists(os.path.join(root_dir, '.git')):
             try:
                 head = subprocess.check_output(['git', 'show', '--no-patch', '--oneline'],
                                                cwd=root_dir)
                 head = head.decode('UTF-8').strip()
-            except OSError:
+            except Exception as e:
+                adtlog.debug(str(e))
                 head = 'cannot determine current HEAD'
             adtlog.info('git checkout: %s' % head)
         else:
@@ -138,7 +150,7 @@ class Testbed:
             self.sp.stdin.close()
             ec = self.sp.wait()
         if ec:
-            self.command('auxverb_debug_fail')
+            self.debug_fail()
             self.bomb('testbed gave exit status %d after quit' % ec)
         self.sp = None
 
@@ -156,27 +168,40 @@ class Testbed:
         # usually "noexec" and /[s]bin might be readonly, so create in /tmp
         if 'reboot' in self.caps and 'root-on-testbed' in self.caps:
             adtlog.debug('testbed supports reboot, creating /tmp/autopkgtest-reboot')
-            self.execute(['sh', '-ecC', '''[ ! -e /tmp/autopkgtest-reboot ] || exit 0; '''
-                          '''/bin/echo -e '#!/bin/sh -e\\n'''
-                          '''[ -n "$1" ] || { echo "Usage: $0 <mark>" >&2; exit 1; }\\n'''
-                          '''echo "$1" > /run/autopkgtest-reboot-mark\\n'''
-                          '''test_script_pid=$(cat /tmp/autopkgtest_script_pid)\\n'''
-                          '''p=$PPID; while true; do read _ c _ pp _ < /proc/$p/stat;'''
-                          '''  [ $pp -ne $test_script_pid ] || break; p=$pp; done\\n'''
-                          '''kill -KILL $p\\n' > /tmp/autopkgtest-reboot;'''
-                          '''chmod 755 /tmp/autopkgtest-reboot;'''
-                          '''[ -L /sbin/autopkgtest-reboot ] || ln -s '''
-                          '''  /tmp/autopkgtest-reboot /sbin/autopkgtest-reboot 2>/dev/null || true'''])
+            reboot = Path(
+                self,
+                os.path.join(PKGDATADIR, 'lib', 'in-testbed', 'reboot.sh'),
+                '%s/autopkgtest-reboot' % self.scratch,
+                False,
+            )
+            reboot.copydown(mode='0755')
+            # TODO: Because we're now creating a symlink, it would be OK
+            # to put the symlink in /run again, even though it's noexec,
+            # because the actual content is elsewhere
+            self.check_exec(
+                ['ln', '-fns', reboot.tb, '/tmp/autopkgtest-reboot'],
+            )
+            # Intentionally not check_exec, and intentionally silencing stderr:
+            # /[s]bin might be read-only.
+            self.execute(
+                ['ln', '-fns', reboot.tb, '/sbin/autopkgtest-reboot'],
+                stderr=subprocess.PIPE,
+            )
 
-            self.execute(['sh', '-ecC', '''[ ! -e /tmp/autopkgtest-reboot-prepare ] || exit 0; '''
-                          '''/bin/echo -e '#!/bin/sh -e\\n'''
-                          '''[ -n "$1" ] || { echo "Usage: $0 <mark>" >&2; exit 1; }\\n'''
-                          '''echo "$1" > /run/autopkgtest-reboot-prepare-mark\\n'''
-                          '''test_script_pid=$(cat /tmp/autopkgtest_script_pid)\\n'''
-                          '''kill -KILL $test_script_pid\\n'''
-                          '''while [ -e /run/autopkgtest-reboot-prepare-mark ]; do sleep 0.5; done\\n'''
-                          ''' '> /tmp/autopkgtest-reboot-prepare;'''
-                          '''chmod 755 /tmp/autopkgtest-reboot-prepare;'''])
+            reboot_prepare = Path(
+                self,
+                os.path.join(PKGDATADIR, 'lib', 'in-testbed',
+                             'reboot-prepare.sh'),
+                '%s/autopkgtest-reboot-prepare' % self.scratch,
+                False,
+            )
+            reboot_prepare.copydown(mode='0755')
+            self.check_exec(
+                [
+                    'ln', '-fns', reboot_prepare.tb,
+                    '/tmp/autopkgtest-reboot-prepare',
+                ],
+            )
 
         # record running kernel version
         kver = self.check_exec(['uname', '-srv'], True).strip()
@@ -221,6 +246,7 @@ class Testbed:
                     self.bomb('testbed boot setup commands failed with status %i' % rc)
 
     def _opened(self, pl):
+        self._tried_provide_sudo = False
         self.scratch = pl[0]
         self.deps_installed = []
         self.apt_pin_for_releases = []
@@ -233,6 +259,12 @@ class Testbed:
             if c.startswith('downtmp-host='):
                 self.shared_downtmp = c.split('=', 1)[1]
 
+        Path(
+            self, os.path.join(PKGDATADIR, 'lib', 'in-testbed', 'wrapper.sh'),
+            '%s/wrapper.sh' % self.scratch,
+            False,
+        ).copydown(mode='0755')
+
         # provide a default for --user
         if self.user is None and 'root-on-testbed' in self.caps:
             self.user = ''
@@ -240,6 +272,30 @@ class Testbed:
                 if c.startswith('suggested-normal-user='):
                     self.user = c.split('=', 1)[1]
 
+            if 'revert-full-system' in self.caps and not self.user:
+                with open(
+                    os.path.join(PKGDATADIR, 'setup-commands', 'create-normal-user'),
+                    encoding='UTF-8',
+                ) as reader:
+                    self.check_exec(['sh', '-euc', reader.read()])
+                    self._created_user = True
+                    self.user = self.check_exec(
+                        ['cat', '/run/autopkgtest-normal-user'],
+                        stdout=True,
+                    ).strip()
+        elif self._created_user:
+            # Re-create the same user after reset. We only do this if
+            # we created it before the reset, in which case the capabilities
+            # should be the same as above and the same username should still
+            # be available.
+            assert 'revert-full-system' in self.caps, self.caps
+            assert 'root-on-testbed' in self.caps, self.caps
+            with open(
+                os.path.join(PKGDATADIR, 'setup-commands', 'create-normal-user'),
+                encoding='UTF-8',
+            ) as reader:
+                self.check_exec(['sh', '-euc', reader.read(), 'sh', self.user])
+
         self.run_setup_commands()
 
         # determine testbed architecture
@@ -420,6 +476,100 @@ class Testbed:
             return
         self.satisfy_dependencies_string(', '.join(deps_new), 'install-deps', shell_on_failure=shell_on_failure, synth_deps=synth_deps)
 
+    def _provide_sudo(self) -> str:
+        '''
+        Enable the needs-sudo restriction, or return why we can't.
+        '''
+
+        if 'root-on-testbed' in self.caps and not self.user:
+            return (
+                'Cannot enable needs-sudo restriction: no ordinary user '
+                'available'
+            )
+
+        # This is intentionally similar to how run_test() runs it, to make
+        # sure that we can sudo here if and only if we could sudo in the test.
+        if 'root-on-testbed' in self.caps:
+            run_as_user = ['su', '-s', '/bin/bash', self.user, '-c']
+        else:
+            run_as_user = ['bash', '-c']
+
+        # First check whether the test user is in the sudo group
+        rc, out, err = self.execute(
+            run_as_user + ['id -Gn'],
+            stdout=subprocess.PIPE,
+            stderr=subprocess.PIPE,
+        )
+
+        if rc == 0 and 'sudo' in out.split():
+            adtlog.debug('User "%s" groups include sudo (%s)'
+                         % (self.user, out.strip()))
+            # Also check whether they need a password (by default yes,
+            # but in e.g. Ubuntu cloud images they don't)
+            rc, out, err = self.execute(
+                run_as_user + ['sudo -n -v'],
+                stdout=subprocess.PIPE,
+                stderr=subprocess.PIPE,
+            )
+            adtlog.debug('sudo -n -v stdout: %s' % out.strip())
+            adtlog.debug('sudo -n -v stderr: %s' % err.strip())
+            adtlog.debug('sudo -n -v status: %d' % rc)
+            already_has_sudo = (rc == 0)
+        else:
+            adtlog.debug('id -Gn stdout: %s' % out.strip())
+            adtlog.debug('id -Gn stderr: %s' % err.strip())
+            adtlog.debug('id -Gn status: %d' % rc)
+            already_has_sudo = False
+
+        if already_has_sudo:
+            adtlog.debug('User "%s" can already sudo without password' %
+                         self.user)
+            return ''
+
+        if 'root-on-testbed' not in self.caps:
+            return 'Cannot enable needs-sudo restriction: not root'
+
+        if 'revert-full-system' not in self.caps:
+            return 'Cannot enable needs-sudo restriction: cannot revert'
+
+        self.needs_reset()
+        adtlog.info('Setting up user "%s" to sudo without password...' %
+                    self.user)
+
+        with open(
+            os.path.join(PKGDATADIR, 'setup-commands', 'enable-sudo')
+        ) as reader:
+            script = reader.read()
+
+        rc, _, _ = self.execute(
+            ['sh', '-euc', script, 'enable-sudo', self.user],
+        )
+
+        if rc != 0:
+            return (
+                'Failed to enable needs-sudo restriction: exit '
+                'status %d'
+            ) % rc
+
+        return ''
+
+    def satisfy_restrictions(
+        self,
+        name: str,
+        restrictions: Set[str],
+    ) -> None:
+        '''
+        Try to satisfy restrictions that can be set up dynamically.
+        Raise Unsupported if unable to do so.
+        '''
+
+        if 'needs-sudo' in restrictions:
+            if not self._tried_provide_sudo:
+                self._tried_provide_sudo = self._provide_sudo() or True
+
+            if self._tried_provide_sudo is not True:
+                raise Unsupported(name, self._tried_provide_sudo)
+
     def needs_reset(self):
         # show what caused a reset
         (fname, lineno, function, code) = traceback.extract_stack(limit=2)[0]
@@ -437,7 +587,22 @@ class Testbed:
         adtlog.debug('%s %s' % (_type.__name__, m))
         raise _type(m)
 
-    def send(self, string, debug_if_fails=True):
+    def debug_fail(self):
+        # Avoid recursion, and also avoid repeating ourselves if cleanup
+        # fails after we already tried to debug a failure
+        if self._tried_debug:
+            return
+
+        self._tried_debug = True
+
+        # Many reasons for failure will make auxverb_debug_fail also fail.
+        # Ignore that, to preserve the original exception
+        try:
+            self.command('auxverb_debug_fail')
+        except Exception as e:
+            adtlog.debug('auxverb_debug_fail failed: %s' % e)
+
+    def send(self, string):
         try:
             adtlog.debug('sending command to testbed: ' + string)
             self.sp.stdin.write(string)
@@ -445,16 +610,13 @@ class Testbed:
             self.sp.stdin.flush()
             self.lastsend = string
         except Exception as e:
-            # command() calls back into send() - avoid potential infinite
-            # recursion
-            if debug_if_fails:
-                self.command('auxverb_debug_fail', debug_if_fails=False)
+            self.debug_fail()
             self.bomb('cannot send to testbed: %s' % e)
 
     def expect(self, keyword, nresults):
         line = self.sp.stdout.readline()
         if not line:
-            self.command('auxverb_debug_fail')
+            self.debug_fail()
             self.bomb('unexpected eof from the testbed')
         if not line.endswith('\n'):
             self.bomb('unterminated line from the testbed')
@@ -464,12 +626,12 @@ class Testbed:
         if not ll:
             self.bomb('unexpected whitespace-only line from the testbed')
         if ll[0] != keyword:
+            self.debug_fail()
+
             if self.lastsend is None:
-                self.command('auxverb_debug_fail')
                 self.bomb("got banner `%s', expected `%s...'" %
                           (line, keyword))
             else:
-                self.command('auxverb_debug_fail')
                 self.bomb("sent `%s', got `%s', expected `%s...'" %
                           (self.lastsend, line, keyword))
         ll = ll[1:]
@@ -479,7 +641,7 @@ class Testbed:
                       (self.lastsend, line, len(ll), nresults))
         return ll
 
-    def command(self, cmd, args=(), nresults=0, unquote=True, debug_if_fails=True):
+    def command(self, cmd, args=(), nresults=0, unquote=True):
         # pass args=[None,...] or =(None,...) to avoid more url quoting
         if type(cmd) is str:
             cmd = [cmd]
@@ -488,7 +650,7 @@ class Testbed:
         else:
             args = list(map(urllib.parse.quote, args))
         al = cmd + args
-        self.send(' '.join(al), debug_if_fails=debug_if_fails)
+        self.send(' '.join(al))
         ll = self.expect('ok', nresults)
         if unquote:
             ll = list(map(urllib.parse.unquote, ll))
@@ -509,7 +671,6 @@ class Testbed:
             env.append('DEBIAN_FRONTEND=noninteractive')
             env.append('APT_LISTBUGS_FRONTEND=none')
             env.append('APT_LISTCHANGES_FRONTEND=none')
-        env += self.install_tmp_env
 
         adtlog.debug('testbed command %s, kind %s, sout %s, serr %s, env %s' %
                      (argv, kind, stdout and 'pipe' or 'raw',
@@ -542,13 +703,13 @@ class Testbed:
                 adtlog.error(msg)
                 raise
             else:
-                self.command('auxverb_debug_fail')
+                self.debug_fail()
                 self.bomb(msg)
 
         adtlog.debug('testbed command exited with code %i' % proc.returncode)
 
         if proc.returncode in (254, 255):
-            self.command('auxverb_debug_fail')
+            self.debug_fail()
             self.bomb('testbed auxverb failed with exit code %i' % proc.returncode)
 
         return (proc.returncode, out, err)
@@ -692,6 +853,7 @@ Description: satisfy autopkgtest test dependencies
                             continue
                         if not installed_virtual_packages:
                             adtlog.warning('no alternative in %s is installed though one should be' % dep)
+                            continue
                         if len(installed_virtual_packages) > 1:
                             adtlog.warning('more than one test dependency alternative in %s installed as a virtual package, installing the first one (%s) as the real package' % (dep, installed_virtual_packages[0]))
                         need_explicit_install.append(installed_virtual_packages[0])
@@ -779,101 +941,6 @@ Description: satisfy autopkgtest test dependencies
 
         self.execute(['dpkg', '--purge', 'autopkgtest-satdep'])
 
-    def install_click(self, clickpath):
-        # copy click into testbed
-        tp = Path(self, clickpath, os.path.join(
-            self.scratch, os.path.basename(clickpath)))
-        tp.copydown()
-        # install it
-        clickopts = ['--all-users']
-        if 'AUTOPKGTEST_CLICK_NO_FRAMEWORK_CHECK' in os.environ:
-            # this is mostly for testing
-            clickopts.append('--force-missing-framework')
-        if 'root-on-testbed' in self.caps:
-            rc = self.execute(['click', 'install', '--allow-unauthenticated'] +
-                              clickopts + [tp.tb], kind='install')[0]
-        else:
-            rc = self.execute(['pkcon', 'install-local', '--allow-untrusted',
-                               tp.tb], kind='install')[0]
-        if rc != 0:
-            self.badpkg('click install failed with status %i' % rc)
-
-        # work around https://launchpad.net/bugs/1333215
-        # we don't want su -l here which resets the environment from
-        # self.execute(); so emulate the parts that we want
-        # FIXME: move "run as user" as an argument of execute()/check_exec() and run with -l
-        self.check_exec(['su', '--shell=/bin/sh', self.su_user, '-c',
-                         ('export USER=%s;' % self.user) +
-                         '. /etc/profile >/dev/null 2>&1 || true; '
-                         ' . ~/.profile >/dev/null 2>&1 || true; '
-                         '[ -z "$UPSTART_SESSION" ] || /sbin/initctl --user start click-user-hooks'])
-
-    def apparmor_click(self, clickpkgs, installed_clicks):
-        '''Update AppArmor rules for click tests
-
-        Return True if anything was modified and apparmor_restore_click()
-        needs to be called.
-        '''
-        # check if we are in a click+AppArmor environment
-        if self.execute(['sh', '-ec',
-                         '[ -d /var/cache/apparmor -a -d /var/lib/apparmor/clicks -a ! -e /var/cache/apparmor/click-ap.rules ] && '
-                         'type aa-clickhook >/dev/null 2>&1'])[0] != 0:
-            adtlog.debug('testbed does not have AppArmor/click or already has Autopilot click rules, no need to adjust rules')
-            return False
-        adtlog.debug('testbed has AppArmor/click')
-
-        if 'root-on-testbed' not in self.caps:
-            adtlog.warning('Cannot adjust AppArmor rules without root/sudo '
-                           'privileges; Autopilot tests will fail and test '
-                           'dependencies will not be available!')
-            return False
-
-        rules = 'dbus (receive, send) bus=session path=/com/canonical/Autopilot/**,'
-        for e in self.install_tmp_env:
-            if e.startswith('QT_PLUGIN_PATH='):
-                for p in e.split('=', 1)[1].split(':'):
-                    p = p.strip()
-                    if p:
-                        rules += ' %s/** r,' % p
-                break
-
-        script = '''echo '%s' > /var/cache/apparmor/click-ap.rules; ''' % rules
-
-        if clickpkgs or installed_clicks:
-            adtlog.info('Updating AppArmor rules to allow autopilot introspection for tested clicks')
-            script += 'for c in %s; do ' \
-                '   info=$(click info %s %s/$(basename "$c")); ' \
-                ''' name=$(echo "$info" | sed -rn '/"name"/ {s/^.*: *"([^"]+)",/\\1/; p}'); ''' \
-                ''' version=$(echo "$info" | sed -rn '/"version"/ {s/^.*: *"([^"]+)",/\\1/; p}'); ''' \
-                '   touch -h /var/lib/apparmor/clicks/${name}_*_${version}.json >/dev/null || true; '\
-                'done; ' \
-                'for c in %s; do ' \
-                '    touch -h /var/lib/apparmor/clicks/${c}_*.json 2>/dev/null || true; ' \
-                'done; ' \
-                'aa-clickhook --include=/var/cache/apparmor/click-ap.rules' % (
-                    ' '.join(clickpkgs),
-                    self.user and ('--user ' + self.user) or '',
-                    self.scratch,
-                    ' '.join(installed_clicks))
-        else:
-            adtlog.info('Updating AppArmor rules to allow autopilot introspection for all clicks (will take a minute)...')
-            script += 'aa-clickhook --force --include=/var/cache/apparmor/click-ap.rules'
-
-        if self.execute(['sh', adtlog.verbosity >= 2 and '-exc' or '-ec', script], kind='install')[0] != 0:
-            self.bomb('Failed to update click AppArmor rules')
-
-        return True
-
-    def apparmor_restore_click(self, clickpkgs, installed_clicks):
-        '''Restore AppArmor rules after click tests'''
-
-        adtlog.info('Restoring click package AppArmor rules')
-        # if we only modified some clicks above, --force will be fast, so it's
-        # ok to always do that
-        script = 'rm -f /var/cache/apparmor/click-ap.rules; aa-clickhook --force'
-        if self.execute(['sh', adtlog.verbosity >= 2 and '-exc' or '-ec', script], kind='install')[0] != 0:
-            self.bomb('Failed to update click AppArmor rules')
-
     def satisfy_dependencies_string(self, deps, what,
                                     build_dep=False, shell_on_failure=False, synth_deps=[]):
         '''Install dependencies from a string into the testbed'''
@@ -940,7 +1007,7 @@ Description: satisfy autopkgtest test dependencies
         '''Run shell in testbed for debugging tests'''
 
         adtlog.info(' - - - - - - - - - - running shell - - - - - - - - - -')
-        self.command('shell', [cwd or '/'] + self.install_tmp_env + extra_env)
+        self.command('shell', [cwd or '/'] + extra_env)
 
     def run_test(self, tree, test, extra_env=[], shell_on_failure=False,
                  shell=False, build_parallel=None):
@@ -956,10 +1023,6 @@ Description: satisfy autopkgtest test dependencies
         if test.path and not os.path.exists(os.path.join(tree.host, test.path)):
             self.badpkg('%s does not exist' % test.path)
 
-        for c in test.clicks:
-            self.install_click(c)
-        need_click_restore = self.apparmor_click(test.clicks, test.installed_clicks)
-
         # record installed package versions
         if self.output_dir and self.execute(
             ['sh', '-ec', 'command -v dpkg-query'],
@@ -992,6 +1055,7 @@ Description: satisfy autopkgtest test dependencies
         test_artifacts = '%s/%s-artifacts' % (self.scratch, test.name)
         autopkgtest_tmp = '%s/autopkgtest_tmp' % (self.scratch)
         assert self.nproc is not None
+        # TODO: Move more of this into wrapper.sh
         script = 'set -e; ' \
                  'export USER=`id -nu`; ' \
                  '. /etc/profile >/dev/null 2>&1 || true; ' \
@@ -1008,8 +1072,6 @@ Description: satisfy autopkgtest test dependencies
                  'unset LANGUAGE LC_CTYPE LC_NUMERIC LC_TIME LC_COLLATE '\
                  '  LC_MONETARY LC_MESSAGES LC_PAPER LC_NAME LC_ADDRESS '\
                  '  LC_TELEPHONE LC_MEASUREMENT LC_IDENTIFICATION LC_ALL;' \
-                 'rm -f /tmp/autopkgtest_script_pid; set -C; echo $$ > /tmp/autopkgtest_script_pid; set +C; ' \
-                 'trap "rm -f /tmp/autopkgtest_script_pid" EXIT INT QUIT PIPE; '\
                  'cd "$buildtree"; '\
                  % {'t': tree.tb, 'a': test_artifacts, 'tmp': autopkgtest_tmp,
                     'cpu': build_parallel or self.nproc}
@@ -1020,16 +1082,6 @@ Description: satisfy autopkgtest test dependencies
 
         for e in extra_env:
             script += 'export \'%s\'; ' % e
-        # there's no way to tell su to not reset $PATH, for install-tmp mode;
-        # we also need it to amend fixed values in /etc/environment
-        for e in self.install_tmp_env:
-            script += 'export %s; ' % e
-        # if we have an user upstart session, poke the environment into it
-        if self.install_tmp_env:
-            script += 'if [ -n "$UPSTART_SESSION" ]; then '
-            for e in self.install_tmp_env:
-                script += ' initctl --user set-env "%s"; ' % e
-            script += 'fi; '
 
         if test.path:
             test_cmd = os.path.join(tree.tb, test.path)
@@ -1037,9 +1089,16 @@ Description: satisfy autopkgtest test dependencies
         else:
             test_cmd = "bash -ec '%s'" % test.command.replace("'", "'\\''")
 
-        script += 'touch %(o)s %(e)s; ' \
-                  '%(t)s 2> >(tee -a %(e)s >&2) > >(tee -a %(o)s);' \
-                  % {'t': test_cmd, 'o': so.tb, 'e': se.tb}
+        script += ' '.join([
+            'exec',
+            '{}/wrapper.sh'.format(shlex.quote(self.scratch)),
+            '--script-pid-file=/tmp/autopkgtest_script_pid',
+            '--stderr={}'.format(shlex.quote(se.tb)),
+            '--stdout={}'.format(shlex.quote(so.tb)),
+            '--',
+            test_cmd,
+            ';',
+        ])
 
         if 'needs-root' not in test.restrictions and self.user is not None:
             if 'root-on-testbed' not in self.caps:
@@ -1081,8 +1140,14 @@ Description: satisfy autopkgtest test dependencies
                 timeout = True
                 break
 
-            # did the test invoke autopkgtest-reboot?
-            if os.WIFSIGNALED(rc) and os.WTERMSIG(rc) == signal.SIGKILL and 'reboot' in self.caps:
+            # If the shell or su command that runs wrapper.sh exits with
+            # status 128+SIGKILL, that probably means either wrapper.sh or
+            # the test itself was terminated with SIGKILL, which might be
+            # because the test asked for the testbed to be rebooted.
+            if (
+                rc in (-signal.SIGKILL, 128 + signal.SIGKILL) and
+                'reboot' in self.caps
+            ):
                 adtlog.debug('test process SIGKILLed, checking for reboot marker')
                 (code, reboot_marker, err) = self.execute(
                     ['cat', '/run/autopkgtest-reboot-mark'],
@@ -1197,11 +1262,6 @@ Description: satisfy autopkgtest test dependencies
         # clean up artifacts and AUTOPKGTEST_TMP dirs
         self.check_exec(['rm', '-rf', test_artifacts, autopkgtest_tmp])
 
-        if need_click_restore:
-            self.apparmor_restore_click(test.clicks, test.installed_clicks)
-        else:
-            adtlog.debug('no need to restore click AppArmor profiles')
-
     #
     # helper methods
     #
@@ -1230,13 +1290,17 @@ Description: satisfy autopkgtest test dependencies
                 '''sort -u | tr '\\n' ' ')"; ''' % \
                 ' '.join(srcpkgs)
 
+        # filenames shouldn't contain slashes, so convert them when
+        # they appear in the release name
+        # apt also doesn't like unknown extensions, so replace dot too
+        release_file_name = release.replace('/', '-').replace('.', '_')
         # prefer given packages from series, but make sure that other packages
         # are taken from default release as much as possible
-        script += 'printf "Package: $PKGS\\nPin: release a=%(release)s\\nPin-Priority: 995\\n" > /etc/apt/preferences.d/autopkgtest-%(release)s; ' % \
-            {'release': release}
+        script += 'printf "Package: $PKGS\\nPin: release a=%(release)s\\nPin-Priority: 995\\n" > /etc/apt/preferences.d/autopkgtest-%(release_file_name)s; ' % \
+            {'release': release, 'release_file_name': release_file_name}
         for default in default_releases:
-            script += 'printf "\nPackage: *\\nPin: release a=%(default)s\\nPin-Priority: 990\\n" >> /etc/apt/preferences.d/autopkgtest-%(release)s; ' % \
-                {'release': release, 'default': default}
+            script += 'printf "\nPackage: *\\nPin: release a=%(default)s\\nPin-Priority: 990\\n" >> /etc/apt/preferences.d/autopkgtest-%(release_file_name)s; ' % \
+                {'release_file_name': release_file_name, 'default': default}
         self.check_exec(['sh', '-ec', script])
         self._set_default_release()
         self.apt_pin_for_releases.append(release)
@@ -1308,7 +1372,11 @@ class Path:
         self.tb = tb
         self.is_dir = is_dir
 
-    def copydown(self, check_existing=False):
+    def copydown(
+        self,
+        check_existing=False,
+        mode='',
+    ) -> None:
         '''Copy file from the host to the testbed
 
         If check_existing is True, don't copy if the testbed path already
@@ -1328,7 +1396,9 @@ class Path:
             self.testbed.command('copydown', (self.host, self.tb))
 
         # we usually want our files be readable for the non-root user
-        if self.testbed.user:
+        if mode:
+            self.testbed.check_exec(['chmod', '-R', mode, '--', self.tb])
+        elif self.testbed.user:
             rc = self.testbed.execute(['chown', '-R', self.testbed.user, '--', self.tb],
                                       stderr=subprocess.PIPE)[0]
             if rc != 0:
@@ -1429,3 +1499,14 @@ def killtree(pid):
         os.kill(pid, signal.SIGTERM)
     except OSError:
         pass
+
+
+def locate_setup_command(name: str) -> str:
+    # Shortcut for shipped scripts
+    if '/' not in name:
+        shipped = os.path.join(PKGDATADIR, 'setup-commands', name)
+
+        if os.path.exists(shipped):
+            return shipped
+
+    return name
