@@ -426,7 +426,7 @@ class Qemu:
 
             shutil.copy(data, '%s/efivars.fd' % workdir)
             argv.append('-drive')
-            argv.append('if=pflash,format=raw,unit=0,read-only,file=' + code)
+            argv.append('if=pflash,format=raw,unit=0,read-only=on,file=' + code)
             argv.append('-drive')
             argv.append(
                 'if=pflash,format=raw,unit=1,file=%s/efivars.fd' % workdir
@@ -631,8 +631,7 @@ class Qemu:
             'aarch64': ['aarch64'],
             # According to https://wiki.alpinelinux.org/wiki/S390x/Installation
             's390x': ['s390x'],
-            # According to https://wiki.debian.org/ppc64el/Installation
-            'ppc64el': ['ppc64el'],
+            'ppc64le': ['ppc64le'],
         }.get(uname_m, [])
 
     def get_socket_path(self, name: str) -> str:
